@@ -1,26 +1,33 @@
-"""Tasks for interacting with GCP Cloud Storage"""
+"""Tasks for interacting with GCP Cloud Storage."""
 
 import os
-from functools import partial
 from io import BytesIO
 from pathlib import Path
-from typing import TYPE_CHECKING, Optional, Tuple, Union
+from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union
+from uuid import uuid4
 
-from anyio import to_thread
 from prefect import get_run_logger, task
+from prefect.filesystems import WritableDeploymentStorage, WritableFileSystem
+from prefect.logging import disable_run_logger
+from prefect.utilities.asyncutils import run_sync_in_worker_thread, sync_compatible
+from prefect.utilities.filesystem import filter_files
+from pydantic import Field, validator
+
+# cannot be type_checking only or else `fields = cls.schema()` raises
+# TypeError: issubclass() arg 1 must be a class
+from prefect_gcp.credentials import GcpCredentials
 
 if TYPE_CHECKING:
     from google.cloud.storage import Bucket
 
-    from prefect_gcp.credentials import GcpCredentials
-
 
 @task
 async def cloud_storage_create_bucket(
     bucket: str,
-    gcp_credentials: "GcpCredentials",
+    gcp_credentials: GcpCredentials,
     project: Optional[str] = None,
     location: Optional[str] = None,
+    **create_kwargs: Dict[str, Any],
 ) -> str:
     """
     Creates a bucket.
@@ -31,6 +38,7 @@ async def cloud_storage_create_bucket(
         project: Name of the project to use; overrides the
             gcp_credentials project if provided.
         location: Location of the bucket.
+        **create_kwargs: Additional keyword arguments to pass to `client.create_bucket`.
 
     Returns:
         The bucket name.
@@ -55,22 +63,22 @@ async def cloud_storage_create_bucket(
     logger.info("Creating %s bucket", bucket)
 
     client = gcp_credentials.get_cloud_storage_client(project=project)
-    partial_create_bucket = partial(client.create_bucket, bucket, location=location)
-    await to_thread.run_sync(partial_create_bucket)
+    await run_sync_in_worker_thread(
+        client.create_bucket, bucket, location=location, **create_kwargs
+    )
     return bucket
 
 
 async def _get_bucket(
     bucket: str,
-    gcp_credentials: "GcpCredentials",
+    gcp_credentials: GcpCredentials,
     project: Optional[str] = None,
 ) -> "Bucket":
     """
     Helper function to retrieve a bucket.
     """
     client = gcp_credentials.get_cloud_storage_client(project=project)
-    partial_get_bucket = partial(client.get_bucket, bucket)
-    bucket_obj = await to_thread.run_sync(partial_get_bucket)
+    bucket_obj = await run_sync_in_worker_thread(client.get_bucket, bucket)
     return bucket_obj
 
 
@@ -78,11 +86,12 @@ async def _get_bucket(
 async def cloud_storage_download_blob_as_bytes(
     bucket: str,
     blob: str,
-    gcp_credentials: "GcpCredentials",
+    gcp_credentials: GcpCredentials,
     chunk_size: Optional[int] = None,
     encryption_key: Optional[str] = None,
     timeout: Union[float, Tuple[float, float]] = 60,
     project: Optional[str] = None,
+    **download_kwargs: Dict[str, Any],
 ) -> bytes:
     """
     Downloads a blob as bytes.
@@ -91,8 +100,6 @@ async def cloud_storage_download_blob_as_bytes(
         bucket: Name of the bucket.
         blob: Name of the Cloud Storage blob.
         gcp_credentials: Credentials to use for authentication with GCP.
-        path: If provided, downloads the contents to the provided file path;
-            if the path is a directory, automatically joins the blob name.
         chunk_size (int, optional): The size of a chunk of data whenever
             iterating (in bytes). This must be a multiple of 256 KB
             per the API specification.
@@ -102,6 +109,8 @@ async def cloud_storage_download_blob_as_bytes(
             (connect_timeout, read_timeout).
         project: Name of the project to use; overrides the
             gcp_credentials project if provided.
+        **download_kwargs: Additional keyword arguments to pass to
+            `Blob.download_as_bytes`.
 
     Returns:
         A bytes or string representation of the blob object.
@@ -131,8 +140,9 @@ async def cloud_storage_download_blob_as_bytes(
         blob, chunk_size=chunk_size, encryption_key=encryption_key
     )
 
-    partial_download = partial(blob_obj.download_as_bytes, timeout=timeout)
-    contents = await to_thread.run_sync(partial_download)
+    contents = await run_sync_in_worker_thread(
+        blob_obj.download_as_bytes, timeout=timeout, **download_kwargs
+    )
     return contents
 
 
@@ -141,11 +151,12 @@ async def cloud_storage_download_blob_to_file(
     bucket: str,
     blob: str,
     path: Union[str, Path],
-    gcp_credentials: "GcpCredentials",
+    gcp_credentials: GcpCredentials,
     chunk_size: Optional[int] = None,
     encryption_key: Optional[str] = None,
     timeout: Union[float, Tuple[float, float]] = 60,
     project: Optional[str] = None,
+    **download_kwargs: Dict[str, Any],
 ) -> Union[str, Path]:
     """
     Downloads a blob to a file path.
@@ -165,6 +176,8 @@ async def cloud_storage_download_blob_to_file(
             (connect_timeout, read_timeout).
         project: Name of the project to use; overrides the
             gcp_credentials project if provided.
+        **download_kwargs: Additional keyword arguments to pass to
+            `Blob.download_to_filename`.
 
     Returns:
         The path to the blob object.
@@ -203,8 +216,9 @@ async def cloud_storage_download_blob_to_file(
         else:
             path = os.path.join(path, blob)  # keep as str if a str is passed
 
-    partial_download = partial(blob_obj.download_to_filename, path, timeout=timeout)
-    await to_thread.run_sync(partial_download)
+    await run_sync_in_worker_thread(
+        blob_obj.download_to_filename, path, timeout=timeout, **download_kwargs
+    )
     return path
 
 
@@ -213,12 +227,13 @@ async def cloud_storage_upload_blob_from_string(
     data: Union[str, bytes],
     bucket: str,
     blob: str,
-    gcp_credentials: "GcpCredentials",
+    gcp_credentials: GcpCredentials,
     content_type: Optional[str] = None,
     chunk_size: Optional[int] = None,
     encryption_key: Optional[str] = None,
     timeout: Union[float, Tuple[float, float]] = 60,
     project: Optional[str] = None,
+    **upload_kwargs: Dict[str, Any],
 ) -> str:
     """
     Uploads a blob from a string or bytes representation of data.
@@ -229,7 +244,7 @@ async def cloud_storage_upload_blob_from_string(
         blob: Name of the Cloud Storage blob.
         gcp_credentials: Credentials to use for authentication with GCP.
         content_type: Type of content being uploaded.
-        chunk_size (int, optional): The size of a chunk of data whenever
+        chunk_size: The size of a chunk of data whenever
             iterating (in bytes). This must be a multiple of 256 KB
             per the API specification.
         encryption_key: An encryption key.
@@ -238,6 +253,8 @@ async def cloud_storage_upload_blob_from_string(
             (connect_timeout, read_timeout).
         project: Name of the project to use; overrides the
             gcp_credentials project if provided.
+        **upload_kwargs: Additional keyword arguments to pass to
+            `Blob.upload_from_string`.
 
     Returns:
         The blob name.
@@ -268,10 +285,13 @@ async def cloud_storage_upload_blob_from_string(
         blob, chunk_size=chunk_size, encryption_key=encryption_key
     )
 
-    partial_upload = partial(
-        blob_obj.upload_from_string, data, content_type=content_type, timeout=timeout
+    await run_sync_in_worker_thread(
+        blob_obj.upload_from_string,
+        data,
+        content_type=content_type,
+        timeout=timeout,
+        **upload_kwargs,
     )
-    await to_thread.run_sync(partial_upload)
     return blob
 
 
@@ -280,11 +300,13 @@ async def cloud_storage_upload_blob_from_file(
     file: Union[str, Path, BytesIO],
     bucket: str,
     blob: str,
-    gcp_credentials: "GcpCredentials",
+    gcp_credentials: GcpCredentials,
+    content_type: Optional[str] = None,
     chunk_size: Optional[int] = None,
     encryption_key: Optional[str] = None,
     timeout: Union[float, Tuple[float, float]] = 60,
     project: Optional[str] = None,
+    **upload_kwargs: Dict[str, Any],
 ) -> str:
     """
     Uploads a blob from file path or file-like object. Usage for passing in
@@ -296,7 +318,8 @@ async def cloud_storage_upload_blob_from_file(
         bucket: Name of the bucket.
         blob: Name of the Cloud Storage blob.
         gcp_credentials: Credentials to use for authentication with GCP.
-        chunk_size (int, optional): The size of a chunk of data whenever
+        content_type: Type of content being uploaded.
+        chunk_size: The size of a chunk of data whenever
             iterating (in bytes). This must be a multiple of 256 KB
             per the API specification.
         encryption_key: An encryption key.
@@ -305,6 +328,8 @@ async def cloud_storage_upload_blob_from_file(
             (connect_timeout, read_timeout).
         project: Name of the project to use; overrides the
             gcp_credentials project if provided.
+        **upload_kwargs: Additional keyword arguments to pass to
+            `Blob.upload_from_file` or `Blob.upload_from_filename`.
 
     Returns:
         The blob name.
@@ -336,10 +361,21 @@ async def cloud_storage_upload_blob_from_file(
     )
 
     if isinstance(file, BytesIO):
-        partial_upload = partial(blob_obj.upload_from_file, file, timeout=timeout)
+        await run_sync_in_worker_thread(
+            blob_obj.upload_from_file,
+            file,
+            content_type=content_type,
+            timeout=timeout,
+            **upload_kwargs,
+        )
     else:
-        partial_upload = partial(blob_obj.upload_from_filename, file, timeout=timeout)
-    await to_thread.run_sync(partial_upload)
+        await run_sync_in_worker_thread(
+            blob_obj.upload_from_filename,
+            file,
+            content_type=content_type,
+            timeout=timeout,
+            **upload_kwargs,
+        )
     return blob
 
 
@@ -348,10 +384,11 @@ async def cloud_storage_copy_blob(
     source_bucket: str,
     dest_bucket: str,
     source_blob: str,
-    gcp_credentials: "GcpCredentials",
+    gcp_credentials: GcpCredentials,
     dest_blob: Optional[str] = None,
     timeout: Union[float, Tuple[float, float]] = 60,
     project: Optional[str] = None,
+    **copy_kwargs: Dict[str, Any],
 ) -> str:
     """
     Copies data from one Google Cloud Storage bucket to another,
@@ -368,6 +405,8 @@ async def cloud_storage_copy_blob(
             (connect_timeout, read_timeout).
         project: Name of the project to use; overrides the
             gcp_credentials project if provided.
+        **copy_kwargs: Additional keyword arguments to pass to
+            `Bucket.copy_blob`.
 
     Returns:
         Destination blob name.
@@ -411,13 +450,217 @@ async def cloud_storage_copy_blob(
         dest_blob = source_blob
 
     source_blob_obj = source_bucket_obj.blob(source_blob)
-    partial_copy_blob = partial(
+    await run_sync_in_worker_thread(
         source_bucket_obj.copy_blob,
         blob=source_blob_obj,
         destination_bucket=dest_bucket_obj,
         new_name=dest_blob,
         timeout=timeout,
+        **copy_kwargs,
     )
-    await to_thread.run_sync(partial_copy_blob)
 
     return dest_blob
+
+
+class GcsBucket(WritableDeploymentStorage, WritableFileSystem):
+    """
+    Block used to store data using GCP Cloud Storage Buckets.
+
+    Attributes:
+        bucket: Name of the bucket.
+        gcp_credentials: The credentials to authenticate with GCP.
+        bucket_folder: A default path to a folder within the GCS bucket to use
+            for reading and writing objects.
+
+    Example:
+        Load stored GCP Cloud Storage Bucket:
+        ```python
+        from prefect_gcp import GcsBucket
+        gcp_cloud_storage_bucket_block = GcsBucket.load("BLOCK_NAME")
+        ```
+    """
+
+    _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/4CD4wwbiIKPkZDt4U3TEuW/c112fe85653da054b6d5334ef662bec4/gcp.png?h=250"  # noqa
+    _block_type_name = "GCS Bucket"
+
+    bucket: str = Field(..., description="Name of the bucket.")
+    gcp_credentials: GcpCredentials = Field(
+        default_factory=GcpCredentials,
+        description="The credentials to authenticate with GCP.",
+    )
+    bucket_folder: str = Field(
+        default="",
+        description=(
+            "A default path to a folder within the GCS bucket to use "
+            "for reading and writing objects."
+        ),
+    )
+
+    @validator("bucket_folder", pre=True, always=True)
+    def _bucket_folder_suffix(cls, value):
+        """
+        Ensures that the bucket folder is suffixed with a forward slash.
+        """
+        if value != "" and not value.endswith("/"):
+            value = f"{value}/"
+        return value
+
+    def _resolve_path(self, path: str) -> str:
+        """
+        A helper function used in write_path to join `self.bucket_folder` and `path`.
+
+        Args:
+            path: Name of the key, e.g. "file1". Each object in your
+                bucket has a unique key (or key name).
+
+        Returns:
+            The joined path.
+        """
+        path = path or str(uuid4())
+
+        # If bucket_folder provided, it means we won't write to the root dir of
+        # the bucket. So we need to add it on the front of the path.
+        path = os.path.join(self.bucket_folder, path) if self.bucket_folder else path
+        return path
+
+    @sync_compatible
+    async def get_directory(
+        self, from_path: Optional[str] = None, local_path: Optional[str] = None
+    ) -> None:
+        """
+        Copies a folder from the configured GCS bucket to a local directory.
+        Defaults to copying the entire contents of the block's bucket_folder
+        to the current working directory.
+
+        Args:
+            from_path: Path in GCS bucket to download from. Defaults to the block's
+                configured bucket_folder.
+            local_path: Local path to download GCS bucket contents to.
+                Defaults to the current working directory.
+        """
+        from_path = (
+            self.bucket_folder if from_path is None else self._resolve_path(from_path)
+        )
+
+        if local_path is None:
+            local_path = os.path.abspath(".")
+        else:
+            local_path = os.path.expanduser(local_path)
+
+        project = self.gcp_credentials.project
+        client = self.gcp_credentials.get_cloud_storage_client(project=project)
+
+        blobs = await run_sync_in_worker_thread(
+            client.list_blobs, self.bucket, prefix=from_path
+        )
+        for blob in blobs:
+            blob_path = blob.name
+            if blob_path[-1] == "/":
+                # object is a folder and will be created if it contains any objects
+                continue
+            local_file_path = os.path.join(local_path, blob_path)
+            os.makedirs(os.path.dirname(local_file_path), exist_ok=True)
+
+            with disable_run_logger():
+                await cloud_storage_download_blob_to_file.fn(
+                    bucket=self.bucket,
+                    blob=blob_path,
+                    path=local_file_path,
+                    gcp_credentials=self.gcp_credentials,
+                )
+
+    @sync_compatible
+    async def put_directory(
+        self,
+        local_path: Optional[str] = None,
+        to_path: Optional[str] = None,
+        ignore_file: Optional[str] = None,
+    ) -> int:
+        """
+        Uploads a directory from a given local path to the configured GCS bucket in a
+        given folder.
+
+        Defaults to uploading the entire contents the current working directory to the
+        block's bucket_folder.
+
+        Args:
+            local_path: Path to local directory to upload from.
+            to_path: Path in GCS bucket to upload to. Defaults to block's configured
+                bucket_folder.
+            ignore_file: Path to file containing gitignore style expressions for
+                filepaths to ignore.
+
+        Returns:
+            The number of files uploaded.
+        """
+        if local_path is None:
+            local_path = os.path.abspath(".")
+        else:
+            local_path = os.path.expanduser(local_path)
+
+        to_path = self.bucket_folder if to_path is None else self._resolve_path(to_path)
+
+        included_files = None
+        if ignore_file:
+            with open(ignore_file, "r") as f:
+                ignore_patterns = f.readlines()
+            included_files = filter_files(local_path, ignore_patterns)
+
+        uploaded_file_count = 0
+        for local_file_path in Path(local_path).rglob("*"):
+            if (
+                included_files is not None
+                and local_file_path.name not in included_files
+            ):
+                continue
+            elif not local_file_path.is_dir():
+                remote_file_path = os.path.join(
+                    to_path, local_file_path.relative_to(local_path)
+                )
+                local_file_content = local_file_path.read_bytes()
+                await self.write_path(remote_file_path, content=local_file_content)
+                uploaded_file_count += 1
+
+        return uploaded_file_count
+
+    @sync_compatible
+    async def read_path(self, path: str) -> bytes:
+        """
+        Read specified path from GCS and return contents. Provide the entire
+        path to the key in GCS.
+
+        Args:
+            path: Entire path to (and including) the key.
+
+        Returns:
+            A bytes or string representation of the blob object.
+        """
+        path = self._resolve_path(path)
+        with disable_run_logger():
+            contents = await cloud_storage_download_blob_as_bytes.fn(
+                bucket=self.bucket, blob=path, gcp_credentials=self.gcp_credentials
+            )
+        return contents
+
+    @sync_compatible
+    async def write_path(self, path: str, content: bytes) -> str:
+        """
+        Writes to an GCS bucket.
+
+        Args:
+            path: The key name. Each object in your bucket has a unique
+                key (or key name).
+            content: What you are uploading to GCS Bucket.
+
+        Returns:
+            The path that the contents were written to.
+        """
+        path = self._resolve_path(path)
+        with disable_run_logger():
+            await cloud_storage_upload_blob_from_string.fn(
+                data=content,
+                bucket=self.bucket,
+                blob=path,
+                gcp_credentials=self.gcp_credentials,
+            )
+        return path
