@@ -1,12 +1,12 @@
-"""Module handling GCP credentials"""
+"""Module handling GCP credentials."""
 
 import functools
-import os
 from pathlib import Path
-from typing import Dict, Optional
+from typing import Dict, Optional, Union
 
+import google.auth.transport.requests
 from google.oauth2.service_account import Credentials
-from pydantic import Json
+from pydantic import Json, root_validator, validator
 
 try:
     from google.cloud.bigquery import Client as BigQueryClient
@@ -24,11 +24,13 @@ except ModuleNotFoundError:
     pass
 
 from prefect.blocks.core import Block
+from prefect.utilities.asyncutils import run_sync_in_worker_thread
 
 
 def _raise_help_msg(key: str):
     """
     Raises a helpful error message.
+
     Args:
         key: the key to access HELP_URLS
     """
@@ -66,103 +68,141 @@ class GcpCredentials(Block):
     [Authentication docs](https://cloud.google.com/docs/authentication/production)
     for more info about the possible credential configurations.
 
-    Args:
+    Attributes:
         service_account_file: Path to the service account JSON keyfile.
-        service_account_info: The contents of the keyfile as a JSON string / dictionary.
-        project: Name of the project to use.
+        service_account_info: The contents of the keyfile as a dict or JSON string.
+
+    Example:
+        Load stored GCP credentials:
+        ```python
+        from prefect_gcp import GcpCredentials
+        gcp_credentials_block = GcpCredentials.load("BLOCK_NAME")
+        ```
     """
 
     _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/4CD4wwbiIKPkZDt4U3TEuW/c112fe85653da054b6d5334ef662bec4/gcp.png?h=250"  # noqa
     _block_type_name = "GCP Credentials"
 
     service_account_file: Optional[Path] = None
-    service_account_info: Optional[Json] = None
+    service_account_info: Optional[Union[Dict[str, str], Json]] = None
     project: Optional[str] = None
 
-    @staticmethod
-    def _get_credentials_from_service_account(
-        service_account_file: Optional[str] = None,
-        service_account_info: Optional[Dict[str, str]] = None,
-    ) -> Credentials:
+    @root_validator
+    def _provide_one_service_account_source(cls, values):
         """
-        Helper method to serialize credentials by using either
-        service_account_file or service_account_info.
+        Ensure that only a service account file or service account info ias provided.
         """
-        if service_account_info and service_account_file:
+        if (
+            values.get("service_account_info") is not None
+            and values.get("service_account_file") is not None
+        ):
             raise ValueError(
                 "Only one of service_account_info or service_account_file "
                 "can be specified at once"
             )
-        elif service_account_file:
-            if not os.path.exists(service_account_file):
-                raise ValueError("The provided path to the service account is invalid")
-            elif isinstance(service_account_file, Path):
-                service_account_file = service_account_file.expanduser()
-            else:
-                service_account_file = os.path.expanduser(service_account_file)
-            credentials = Credentials.from_service_account_file(service_account_file)
-        elif service_account_info:
-            credentials = Credentials.from_service_account_info(service_account_info)
+        return values
+
+    @validator("service_account_file")
+    def _check_service_account_file(cls, file):
+        """Get full path of provided file and make sure that it exists."""
+        if not file:
+            return file
+
+        service_account_file = Path(file).expanduser()
+        if not service_account_file.exists():
+            raise ValueError("The provided path to the service account is invalid")
+        return service_account_file
+
+    def block_initialization(self):
+        if self.project is None and (
+            self.service_account_file or self.service_account_info
+        ):
+            credentials = self.get_credentials_from_service_account()
+            self.project = credentials.project
+
+    def get_credentials_from_service_account(self) -> Union[Credentials, None]:
+        """
+        Helper method to serialize credentials by using either
+        service_account_file or service_account_info.
+        """
+        if self.service_account_file:
+            credentials = Credentials.from_service_account_file(
+                self.service_account_file,
+                scopes=["https://www.googleapis.com/auth/cloud-platform"],
+            )
+        elif self.service_account_info:
+            credentials = Credentials.from_service_account_info(
+                self.service_account_info,
+                scopes=["https://www.googleapis.com/auth/cloud-platform"],
+            )
         else:
             return None
         return credentials
 
+    async def get_access_token(self):
+        """
+        See: https://stackoverflow.com/a/69107745
+        Also: https://www.jhanley.com/google-cloud-creating-oauth-access-tokens-for-rest-api-calls/
+        """  # noqa
+        request = google.auth.transport.requests.Request()
+        credentials = self.get_credentials_from_service_account()
+        await run_sync_in_worker_thread(credentials.refresh, request)
+
+        return credentials.token
+
     @_raise_help_msg("cloud_storage")
     def get_cloud_storage_client(
         self, project: Optional[str] = None
     ) -> "StorageClient":
         """
+        Gets an authenticated Cloud Storage client.
+
         Args:
             project: Name of the project to use; overrides the base
                 class's project if provided.
 
+        Returns:
+            An authenticated Cloud Storage client.
+
         Examples:
             Gets a GCP Cloud Storage client from a path.
             ```python
             from prefect import flow
             from prefect_gcp.credentials import GcpCredentials
-
             @flow()
             def example_get_client_flow():
                 service_account_file = "~/.secrets/prefect-service-account.json"
                 client = GcpCredentials(
                     service_account_file=service_account_file
                 ).get_cloud_storage_client()
-
             example_get_client_flow()
             ```
 
-            Gets a GCP Cloud Storage client from a JSON str.
+            Gets a GCP Cloud Storage client from a dictionary.
             ```python
-            import json
             from prefect import flow
             from prefect_gcp.credentials import GcpCredentials
-
             @flow()
             def example_get_client_flow():
-                service_account_info = json.dumps({
+                service_account_info = {
                     "type": "service_account",
                     "project_id": "project_id",
                     "private_key_id": "private_key_id",
-                    "private_key": private_key",
+                    "private_key": "private_key",
                     "client_email": "client_email",
                     "client_id": "client_id",
                     "auth_uri": "auth_uri",
                     "token_uri": "token_uri",
                     "auth_provider_x509_cert_url": "auth_provider_x509_cert_url",
                     "client_x509_cert_url": "client_x509_cert_url"
-                })
+                }
                 client = GcpCredentials(
                     service_account_info=service_account_info
                 ).get_cloud_storage_client()
-
             example_get_client_flow()
             ```
         """
-        credentials = self._get_credentials_from_service_account(
-            service_account_file=self.service_account_file,
-            service_account_info=self.service_account_info,
-        )
+        credentials = self.get_credentials_from_service_account()
 
         # override class project if method project is provided
         project = project or self.project
@@ -174,58 +214,56 @@ class GcpCredentials(Block):
         self, project: str = None, location: str = None
     ) -> "BigQueryClient":
         """
+        Gets an authenticated BigQuery client.
+
         Args:
             project: Name of the project to use; overrides the base
                 class's project if provided.
             location: Location to use.
 
+        Returns:
+            An authenticated BigQuery client.
+
         Examples:
             Gets a GCP BigQuery client from a path.
             ```python
             from prefect import flow
             from prefect_gcp.credentials import GcpCredentials
-
             @flow()
             def example_get_client_flow():
                 service_account_file = "~/.secrets/prefect-service-account.json"
                 client = GcpCredentials(
                     service_account_file=service_account_file
                 ).get_bigquery_client()
-
             example_get_client_flow()
             ```
 
-            Gets a GCP BigQuery client from a JSON str.
+            Gets a GCP BigQuery client from a dictionary.
             ```python
-            import json
             from prefect import flow
             from prefect_gcp.credentials import GcpCredentials
-
             @flow()
             def example_get_client_flow():
-                service_account_info = json.dumps({
+                service_account_info = {
                     "type": "service_account",
                     "project_id": "project_id",
                     "private_key_id": "private_key_id",
-                    "private_key": private_key",
+                    "private_key": "private_key",
                     "client_email": "client_email",
                     "client_id": "client_id",
                     "auth_uri": "auth_uri",
                     "token_uri": "token_uri",
                     "auth_provider_x509_cert_url": "auth_provider_x509_cert_url",
                     "client_x509_cert_url": "client_x509_cert_url"
-                })
+                }
                 client = GcpCredentials(
                     service_account_info=service_account_info
-                ).get_bigquery_client(json)
+                ).get_bigquery_client()
 
             example_get_client_flow()
             ```
         """
-        credentials = self._get_credentials_from_service_account(
-            service_account_file=self.service_account_file,
-            service_account_info=self.service_account_info,
-        )
+        credentials = self.get_credentials_from_service_account()
 
         # override class project if method project is provided
         project = project or self.project
@@ -237,35 +275,32 @@ class GcpCredentials(Block):
     @_raise_help_msg("secret_manager")
     def get_secret_manager_client(self) -> "SecretManagerServiceClient":
         """
-        Args:
-            project: Name of the project to use; overrides the base
-                class's project if provided.
+        Gets an authenticated Secret Manager Service client.
+
+        Returns:
+            An authenticated Secret Manager Service client.
 
         Examples:
             Gets a GCP Secret Manager client from a path.
             ```python
             from prefect import flow
             from prefect_gcp.credentials import GcpCredentials
-
             @flow()
             def example_get_client_flow():
                 service_account_file = "~/.secrets/prefect-service-account.json"
                 client = GcpCredentials(
                     service_account_file=service_account_file
                 ).get_secret_manager_client()
-
             example_get_client_flow()
             ```
 
-            Gets a GCP Cloud Storage client from a JSON str.
+            Gets a GCP Cloud Storage client from a dictionary.
             ```python
-            import json
             from prefect import flow
             from prefect_gcp.credentials import GcpCredentials
-
             @flow()
             def example_get_client_flow():
-                service_account_info = json.dumps({
+                service_account_info = {
                     "type": "service_account",
                     "project_id": "project_id",
                     "private_key_id": "private_key_id",
@@ -280,14 +315,10 @@ class GcpCredentials(Block):
                 client = GcpCredentials(
                     service_account_info=service_account_info
                 ).get_secret_manager_client()
-
             example_get_client_flow()
             ```
         """
-        credentials = self._get_credentials_from_service_account(
-            service_account_file=self.service_account_file,
-            service_account_info=self.service_account_info,
-        )
+        credentials = self.get_credentials_from_service_account()
 
         # doesn't accept project; must pass in project in tasks
         secret_manager_client = SecretManagerServiceClient(credentials=credentials)
